from .grammar import grammar
from .visitor import JSCNVisitor
from .tree import Type
from typing import Any


__version__ = "0.4"


def to_tree(source: str) -> Type:
    raw_tree = grammar.parse(source)
    visitor = JSCNVisitor()
    parsed_tree = visitor.visit(raw_tree)
    return parsed_tree


def to_schema(source: str) -> Any:
    return to_tree(source).to_schema()


def to_json(source: str) -> str:
    return to_tree(source).to_json()



from abc import ABC, abstractmethod
from typing import NamedTuple, Optional, Set
import json


class Type(ABC):
    def __init__(self, *args, **kwargs):
        self.args = args
        self.kwargs = kwargs

    @abstractmethod
    def to_schema(self):
        pass

    def to_json(self):
        return json.dumps(self.to_schema())

    def __str__(self):
        a = [str(arg) for arg in self.args] + [
            k + "=" + str(v) for k, v in self.kwargs.items()
        ]
        return f"{self.__class__.__name__}({', '.join(a)})"

    def _combine(self, other, op):
        args = []
        for it in (self, other):
            if isinstance(it, Operator) and it.args[0] == op:
                args.extend(it.args[1:])
            else:
                args.append(it)
        return Operator(op, *args)

    def __and__(self, other):
        return self._combine(other, 'allOf')

    def __or__(self, other):
        return self._combine(other, 'anyOf')

    def __not__(self):
        return Not(self)
    __repr__ = __str__


class Entry(Type):
    def to_schema(self, check_definitions=True):
        r = self.args[0].to_schema()
        definitions = self.args[1]
        if definitions:
            r["definitions"] = {k: v.to_schema() for k, v in definitions.items()}
            if check_definitions:
                for k in self.get_references(r):
                    if k not in r["definitions"].keys():
                        raise ValueError(f"Missing definition for {k}")
        r["$schema"] = "http://json-schema.org/draft-07/schema#"
        return r

    def get_references(self, x) -> Set[str]:
        """Extract every definition usage from a schema, so that it can be checked
        that they are all defined."""
        if isinstance(x, dict):
            if "$ref" in x:
                return {x["$ref"].rsplit("/", 1)[-1]}
            else:
                return set.union(*(self.get_references(y) for y in x.values()))
        elif isinstance(x, list):
            return set.union(*(self.get_references(y) for y in x))
        else:
            return set()

    def _combine(self, other, op):
        # TODO: merge definitions and keep the `"$schema"` tag.
        return super().combine(other, op)


class Integer(Type):
    def to_schema(self):
        (card_min, card_max) = self.kwargs["cardinal"]
        mult = self.kwargs["multiple"]
        r = {"type": "integer"}
        if card_min is not None:
            r["minimum"] = card_min
        if card_max is not None:
            r["maximum"] = card_max
        if mult is not None:
            r["multipleOf"] = mult
        return r


class String(Type):
    def to_schema(self):
        card = self.kwargs.get("cardinal", (None, None))
        r = {"type": "string"}
        if card is None:
            return r
        if isinstance(card, int):
            card = (card, card)
        card_min, card_max = card
        if card_min is not None:
            r["minLength"] = card_min
        if card_max is not None:
            r["maxLength"] = card_max
        if "format" in self.kwargs:
            r["format"] = self.kwargs["format"]
        if "regex" in self.kwargs:
            r["pattern"] = self.kwargs["regex"]
        return r


class Litteral(Type):
    def to_schema(self):
        return {"type": self.args[0]}


class Constant(Type):
    def to_schema(self):
        return {"const": self.args[0]}


class Operator(Type):
    def to_schema(self):
        op = self.args[0]
        args = self.args[1:]
        return {op: [a.to_schema() for a in args]}


class Not(Type):
    def __not__(self):
        return self.args[0]
    def to_schema(self):
        return {"not": self.args[0].to_schema()}


class Enum(Type):
    def to_schema(self):
        return {"enum": list(self.args)}


class ObjectProperty(NamedTuple):
    name: Optional[str]
    optional: bool
    type: Type


class Object(Type):
    def to_schema(self):
        # TODO: detect inconsistency between "only" and a "_" property wildcard
        pairs = self.kwargs.get("properties", {})
        card_min, card_max = self.kwargs.get("cardinal", (None, None))
        r = {"type": "object"}
        properties = {}
        required = []
        for (k, opt, v) in pairs:
            if k is None:
                r["additionalProperties"] = v.to_schema()
                continue
            if not opt:
                required.append(k)
            if v is not None:
                properties[k] = v.to_schema()
        if required:
            r["required"] = required
        if properties:
            r["properties"] = properties
        if (
            not self.kwargs.get("additional_properties")
            and "additionalProperties" not in r
        ):
            r["additionalProperties"] = False

        implicit_card_min = len(required)
        implicit_card_max = (
            len(properties) if r.get("additionalProperties") is False else None
        )

        if 'property_names' in self.kwargs:
            r['propertyNames'] = self.kwargs['property_names'].to_schema()
            # TODO it would be neat to accept definitions here

        if card_min is not None:
            if card_min > implicit_card_min:
                r["minProperties"] = card_min
            if implicit_card_max is not None and card_min > implicit_card_max:
                raise ValueError(
                    f"Can only have up to {implicit_card_max} properties, not {card_min}"
                )

        if card_max is not None:
            if implicit_card_max is None or card_max < implicit_card_max:
                r["maxProperties"] = card_max
            if implicit_card_min > card_max:
                raise ValueError(
                    f"Must have at least {implicit_card_min} properties, which is more than {card_max}"
                )
        return r


class Array(Type):
    def to_schema(self):
        types = self.kwargs["types"]
        extra_type = self.kwargs["additional_types"]
        card_min, card_max = self.kwargs["cardinal"]
        r = {"type": "array"}

        if types:  # Tuple array
            r["items"] = [t.to_schema() for t in types]
            if extra_type is False:  # No extra items allowed
                r["additionalItems"] = False
            elif extra_type is True:  # Extra items with any type
                pass
            else:  # Forced type for extra items
                r["additionalItems"] = extra_type.to_schema()
        elif isinstance(extra_type, Type):  # List array with homogeneous type
            r["items"] = extra_type.to_schema()

        implicit_card_min = len(types)
        implicit_card_max = len(types) if extra_type is False else None

        if card_min is not None and card_min > implicit_card_min:
            r["minItems"] = card_min
            if implicit_card_max is not None and card_min > implicit_card_max:
                raise ValueError(
                    f"Can only have up to {implicit_card_max} items, not {card_min}"
                )
        if card_max is not None:
            if implicit_card_min > card_max:
                raise ValueError(
                    f"Must have at least {implicit_card_min} items, which is more than {card_max}"
                )
            if implicit_card_max is None or card_max < implicit_card_max:
                r["maxItems"] = card_max

        if self.kwargs.get("unique"):
            r['uniqueItems'] = True

        return r


class Pointer(Type):
    def to_schema(self):
        return {"$ref": "#/definitions/" + self.args[0]}



from parsimonious import NodeVisitor
from collections import Sequence
from typing import Tuple, Optional, Set, Dict
import json

from . import tree as T


class JSCNVisitor(NodeVisitor):

    WHITESPACE_TOKEN = object()

    SHELL_EXPRESSIONS = {
        "type": 0,
        "card_content": 0,
        "card_1": (0, 0),
        "card_2": (0, -1),
        "object": 0,
        "object_property": 0,
        "object_pair_type": 0,
        "object_pair_name": 0,
        "array": 0,
        "parens": 1,
    }

    @classmethod
    def unspace(cls, sequence, index=None):
        """Remove space tokens from a sequence of visited children.
        if extra integer indexes are passed as args, only keep those."""
        unspaced = tuple(x for x in sequence if x is not cls.WHITESPACE_TOKEN)
        if index is not None:
            return unspaced[index]
        else:
            return unspaced

    def visit__(self, node, c) -> object:
        return self.WHITESPACE_TOKEN

    def visit_entry(self, node, c) -> T.Entry:
        type, definitions = self.unspace(c)
        return T.Entry(type, definitions)

    def visit_sequence_and(self, node, c) -> T.Type:
        first, rest = self.unspace(c)
        if len(rest):
            return T.Operator("allOf", first, *(item[1] for item in rest))
        else:
            return first

    def visit_sequence_or(self, node, c) -> T.Type:
        first, rest = self.unspace(c)
        if len(rest) == 0:
            return first
        args = [first] + [item[1] for item in rest]
        if all(isinstance(a, T.Constant) for a in args):
            # Convert oneof(const(...)...) into enum(...)
            constants = [a.args[0] for a in args]
            return T.Enum(*constants)
        else:
            return T.Operator("oneOf", first, *(item[1] for item in rest))

    def visit_not_type(self, node, c) -> T.Not:
        return T.Not(self.unspace(c, 1))

    def visit_string(self, node, c) -> T.String:
        _, cardinal = self.unspace(c)
        return T.String(cardinal=cardinal)

    def visit_integer(self, node, c) -> T.Integer:
        _, cardinal, multiple = self.unspace(c)
        return T.Integer(cardinal=cardinal, multiple=multiple)

    def visit_litteral(self, node, c) -> T.Litteral:
        # This rule is space-free
        return T.Litteral(node.children[0].text)

    def visit_lit_string(self, node, c) -> str:
        # This rule is space-free
        source = node.text[1:-1]
        unescaped = source.encode("utf-8").decode("unicode_escape")
        return unescaped

    def visit_lit_regex(self, node, c) -> T.String:
        return T.String(regex=node.children[-1].text[1:-1])

    def visit_lit_format(self, node, c) -> T.String:
        return T.String(format=node.children[-1].text[1:-1])

    def visit_opt_multiple(self, node, c) -> Optional[int]:
        uc = self.unspace(c)
        return None if len(uc) == 0 else uc[0][1]

    def visit_opt_cardinal(self, node, c) -> Tuple[Optional[int], Optional[int]]:
        # generic_visit didn't detect card_1, probably due to some node optimization?
        if len(c) == 0:  # Empty cardinal
            return (None, None)
        uc = self.unspace(c[0], 1)
        if isinstance(uc, int):
            return (uc, uc)
        else:
            return uc

    def visit_card_min(self, node, c) -> Tuple[int, None]:
        return (self.unspace(c, 0), None)

    def visit_card_max(self, node, c) -> Tuple[None, int]:
        return (None, self.unspace(c, 2))

    def visit_lit_integer(self, node, c) -> int:
        # This rule is space-free
        text = node.text
        base = 16 if len(text) > 2 and text[1] == "x" else 10
        return int(text, base)

    def visit_constant(self, node, c) -> T.Constant:
        # This rule is space-free
        source = node.text[1:-1]
        try:
            value = json.loads(source)
            return T.Constant(value)
        except json.JSONDecodeError:
            raise ValueError(f"{source} is not a valid JSON fragment")

    def visit_object_keyword(self, node, c) -> T.Object:
        return T.Object(additional_properties=True, cardinal=c[-1])

    def visit_object_empty(self, node, c) -> T.Object:
        kwargs = {}
        _, (only, only_regex), _, kwargs['cardinal'] = self.unspace(c)
        if not only:
            kwargs['additional_properties'] = True
        elif only_regex is not None:
            kwargs['additional_properties'] = True
            kwargs['property_names'] = only_regex
        else:
            kwargs['additional_properties'] = False
        return T.Object(**kwargs)

    def visit_object_non_empty(self, node, c) -> T.Object:
        _, (only, only_regex), first_field, others_with_commas, _, card = self.unspace(c)
        other_fields = (item[1] for item in others_with_commas)
        fields = (first_field, *other_fields)
        kwargs = {'properties': fields, 'cardinal': card}
        if not only:
            kwargs['additional_properties'] = True
        elif only_regex is not None:
            kwargs['additional_properties'] = True
            kwargs['property_names'] = only_regex
        else:
            kwargs['additional_properties'] = False
        return T.Object(**kwargs)

    def visit_object_pair(self, node, c) -> T.ObjectProperty:
        key, question, _, val = self.unspace(c)
        return T.ObjectProperty(key, bool(question), val)

    def visit_object_pair_unquoted_name(self, node, c) -> str:
        return node.text

    def visit_object_unnamed_pair(self, node, c) -> T.ObjectProperty:
        return T.ObjectProperty(None, True, self.unspace(c, 2))

    def visit_object_only(self, node, c) -> Tuple[bool, Optional[str]]:
        """Parse `only`, `only <pattern>` and `only <pattern>,`."""
        if len(c) == 0:  # Empty sequence
            return False, None
        maybe_regex = self.unspace(c[0], 1)
        if len(maybe_regex):
            return True, maybe_regex[0][0]
        else:
            return True, None

    def visit_array_empty(self, node, c) -> T.Array:
        card = self.unspace(c, -1)
        return T.Array(types=[], additional_types=True, cardinal=card)

    def visit_array_non_empty(self, node, c) -> T.Array:
        """
        With only one type and a "+" / "*" suffix, it's an homogeneous list.
        In this case, if there is a "only" qualifier it is ignored.
        With more than one type and a "+" / "*" suffix, it's a tuple type
        with extra items type. Here too, any "only" qualifier will be ignored.
        Without a suffix, it's a tuple type, the "only" qualifier is enforced.
        """
        _, array_prefix, first_type, other_types_with_commas, extra, _, card = self.unspace(c)

        other_types = (t[1] for t in other_types_with_commas)
        types = (first_type, *other_types)

        if extra is None:  # No suffix -> tuple typing
            additional_items = "only" not in array_prefix
        elif extra == "+":
            additional_items = types[-1]
            # Don't remove it from required items, there must be at least one
        else:  # Last type is the type of extra items
            additional_items = types[-1]
            types = types[:-1]
        if isinstance(card[0], int) and len(types) >= card[0]:
            card = (None, card[1])  # Constraint is redundant
        if isinstance(card[1], int) and not additional_items and len(types) < card[1]:
            raise ValueError(
                f"An array cannot be both {len(types)} and <={card[1]} items long"
            )

        return T.Array(types=types, additional_types=additional_items, cardinal=card, unique="unique" in array_prefix)

    def visit_array_prefix(self, node, c) -> Set[str]:
        """Return a set of strings among "unique" and "only"."""
        r = set()
        for word in ("unique", "only"):
            if word in node.text:
                r.add(word)
        return r

    def visit_array_extra(self, node, c) -> str:
        # This rule is space-free
        if len(c) == 0:
            return None
        else:
            return node.children[0].text

    def visit_opt_definitions(self, node, c) -> Dict[str, T.Type]:
        if len(c) == 0:  # Empty definition
            return {}
        _, first_def, other_defs_with_and = self.unspace(c[0])
        other_defs = [d[1] for d in other_defs_with_and]
        items = (first_def, *other_defs)
        return dict(items)

    def visit_definition(self, node, c) -> Tuple[str, T.Type]:
        id, _, type = self.unspace(c)
        return (id, type)

    def visit_def_identifier(self, node, c) -> str:
        return node.text

    def visit_def_pointer(self, node, c) -> T.Pointer:
        return T.Pointer(c[1])

    def generic_visit(self, node, c) -> tuple:
        """ The generic visit method. """
        n = self.SHELL_EXPRESSIONS.get(node.expr_name)
        unspaced_c = self.unspace(c)
        if n is None:
            return unspaced_c
        elif isinstance(n, Sequence):
            return tuple(unspaced_c[i] for i in n)
        elif isinstance(n, int):
            return unspaced_c[n]
        else:
            raise ValueError(f"bad SHELL_EXPRESSIONS for {node.expr_name}")




import sys
import json
from argparse import ArgumentParser

from .grammar import grammar
from .visitor import JSCNVisitor


def main():
    parser = ArgumentParser(description="Convert from a compact DSL into full JSON schema.")
    parser.add_argument('-o', '--output', default="-",
                        help="Output file; defaults to stdout")
    parser.add_argument('-v', '--verbose', action='store_const', const=True, default=False,
                        help="Verbose output")
    parser.add_argument('filename', nargs="?", default="-",
                        help="Input file; use '-' to read from stdin.")
    args = parser.parse_args()

    input = open(args.filename, 'r') if args.filename != "-" else sys.stdin
    output = open(args.output, 'w') if args.output != "-" else sys.stdout
    verbose = "-v" in sys.argv

    source = input.read()
    raw_tree = grammar.parse(source)
    visitor = JSCNVisitor()
    if verbose:
        print("Raw output:", raw_tree)
    parsed_tree = visitor.visit(raw_tree)
    if verbose:
        print("Parsed output:", parsed_tree)
        schema = parsed_tree.to_schema()
        print("Schema:", schema)
        result = json.dumps(schema)
    else:
        result = parsed_tree.to_json()

    output.write(result+"\n")
    output.flush()



import unittest
from . import to_schema, to_json
import json

class TestJSCN(unittest.TestCase):

    def cmp(self, src: str,sch: dict) -> None:
        sch2 = to_schema(src)
        del sch2['$schema']
        self.assertDictEqual(sch2, sch)

    def test_to_json(self):
        self.assertEqual(json.dumps(to_schema("integer")), to_json("integer"))

    def test_simple(self):
        self.cmp("boolean", {"type": "boolean"})
        self.cmp("string", {"type": "string"})
        self.cmp("integer", {"type": "integer"})
        self.cmp("number", {"type": "number"})
        self.cmp("null", {"type": "null"})

    def test_lit_string(self):
        self.cmp('r"foo"', {"type": "string", "pattern": "foo"})
        self.cmp('f"foo"', {"type": "string", "format": "foo"})
        self.cmp('r"foo\\\"bar"', {"type": "string", "pattern": r'foo\"bar'})
        self.cmp('r"foo\\""', {"type": "string", "pattern": r'foo\"'})


    def test_constant(self):
        self.cmp("`123`", {"const": 123})
        self.cmp('`"123"`', {"const": "123"})
        self.cmp('`{"a": 1}`', {"const": {"a": 1}})

    def test_enum(self):
        self.cmp("`1`|`2`|`3`", {"enum": [1, 2, 3]})

    def test_object_empty(self):
        obj = {"type": "object"}
        self.cmp("{}", obj)
        self.cmp("object", obj)

    def test_object_simple(self):
        self.cmp("{only foo: integer}",
                 {"type": "object",
                  "required": ["foo"],
                  "properties":
                  {"foo": {"type": "integer"}},
                  "additionalProperties": False})
        self.cmp("{foo: integer}",
                 {"type": "object",
                  "required": ["foo"],
                  "properties":
                  {"foo": {"type": "integer"}}})
        self.cmp("{foo?: integer}",
                 {"type": "object",
                  "properties":
                  {"foo": {"type": "integer"}}})
        self.cmp('{"foo"?: integer}',
                 {"type": "object",
                  "properties":
                  {"foo": {"type": "integer"}}})
        self.cmp('{_: integer}',
                 {"type": "object",
                  "additionalProperties": {"type": "integer"}})

        # TODO combine wildcard + only: what's the meaning?!

    def test_object_card(self):
        pass

    def test_array_empty(self):
        array = {"type": "array"}
        self.cmp("[]", array)
        self.cmp("array", array)

    def test_object_simple(self):
        integer = {"type": "integer"}
        self.cmp("[only integer]",
                 {"type": "array",
                  "items": [integer],
                  "additionalItems": False})
        self.cmp("[integer]",
                 {"type": "array",
                  "items": [integer]})
        self.cmp("[integer*]",
                 {"type": "array",
                  "items": integer}) # List-notation rather than tuple notation
        self.cmp("[integer+]",  # Maybe encoded either that way or with a cardinal
                 {"type": "array",
                  "items": [integer],
                  "additionalItems": integer})

    def test_array_card(self):
        pass


if __name__ == '__main__':
    unittest.main()
